"""Synchronous network clients for interacting with the XAHL."""
from xahau.asyncio.clients.client import Client
from xahau.asyncio.clients.exceptions import XAHLRequestFailureException
from xahau.asyncio.clients.utils import (
    json_to_response,
    request_to_json_rpc,
    request_to_websocket,
    websocket_to_response,
)
from xahau.clients.json_rpc_client import JsonRpcClient
from xahau.clients.websocket_client import WebsocketClient

__all__ = [
    "Client",
    "JsonRpcClient",
    "request_to_json_rpc",
    "json_to_response",
    "request_to_websocket",
    "XAHLRequestFailureException",
    "websocket_to_response",
    "WebsocketClient",
]



"""Methods for working with XAHL wallets."""
from xahau.asyncio.wallet import XAHLFaucetException
from xahau.wallet.main import Wallet
from xahau.wallet.wallet_generation import generate_faucet_wallet

__all__ = ["Wallet", "generate_faucet_wallet", "XAHLFaucetException"]



"""
Model objects for specific `types of Transactions
<https://xrpl.org/transaction-types.html>`_ in the XAH Ledger.
"""

from xahau.models.transactions.account_delete import AccountDelete
from xahau.models.transactions.account_set import (
    AccountSet,
    AccountSetAsfFlag,
    AccountSetFlag,
    AccountSetFlagInterface,
)
from xahau.models.transactions.amm_bid import AMMBid, AuthAccount
from xahau.models.transactions.amm_create import AMMCreate
from xahau.models.transactions.amm_delete import AMMDelete
from xahau.models.transactions.amm_deposit import (
    AMMDeposit,
    AMMDepositFlag,
    AMMDepositFlagInterface,
)
from xahau.models.transactions.amm_vote import AMMVote
from xahau.models.transactions.amm_withdraw import (
    AMMWithdraw,
    AMMWithdrawFlag,
    AMMWithdrawFlagInterface,
)
from xahau.models.transactions.check_cancel import CheckCancel
from xahau.models.transactions.check_cash import CheckCash
from xahau.models.transactions.check_create import CheckCreate
from xahau.models.transactions.claim_reward import ClaimReward
from xahau.models.transactions.clawback import Clawback
from xahau.models.transactions.deposit_preauth import DepositPreauth
from xahau.models.transactions.did_delete import DIDDelete
from xahau.models.transactions.did_set import DIDSet
from xahau.models.transactions.escrow_cancel import EscrowCancel
from xahau.models.transactions.escrow_create import EscrowCreate
from xahau.models.transactions.escrow_finish import EscrowFinish
from xahau.models.transactions.import_ import Import
from xahau.models.transactions.invoke import Invoke
from xahau.models.transactions.metadata import TransactionMetadata
from xahau.models.transactions.nftoken_accept_offer import NFTokenAcceptOffer
from xahau.models.transactions.nftoken_burn import NFTokenBurn
from xahau.models.transactions.nftoken_cancel_offer import NFTokenCancelOffer
from xahau.models.transactions.nftoken_create_offer import (
    NFTokenCreateOffer,
    NFTokenCreateOfferFlag,
    NFTokenCreateOfferFlagInterface,
)
from xahau.models.transactions.nftoken_mint import (
    NFTokenMint,
    NFTokenMintFlag,
    NFTokenMintFlagInterface,
)
from xahau.models.transactions.offer_cancel import OfferCancel
from xahau.models.transactions.offer_create import (
    OfferCreate,
    OfferCreateFlag,
    OfferCreateFlagInterface,
)
from xahau.models.transactions.oracle_delete import OracleDelete
from xahau.models.transactions.oracle_set import OracleSet
from xahau.models.transactions.payment import Payment, PaymentFlag, PaymentFlagInterface
from xahau.models.transactions.payment_channel_claim import (
    PaymentChannelClaim,
    PaymentChannelClaimFlag,
    PaymentChannelClaimFlagInterface,
)
from xahau.models.transactions.payment_channel_create import PaymentChannelCreate
from xahau.models.transactions.payment_channel_fund import PaymentChannelFund
from xahau.models.transactions.remit import MintURIToken, Remit
from xahau.models.transactions.set_hook import (
    Hook,
    HookGrant,
    HookParameter,
    SetHook,
    SetHookFlag,
    SetHookFlagInterface,
)
from xahau.models.transactions.set_regular_key import SetRegularKey
from xahau.models.transactions.signer_list_set import SignerEntry, SignerListSet
from xahau.models.transactions.ticket_create import TicketCreate
from xahau.models.transactions.transaction import Memo, Signer, Transaction
from xahau.models.transactions.trust_set import (
    TrustSet,
    TrustSetFlag,
    TrustSetFlagInterface,
)
from xahau.models.transactions.uritoken_burn import URITokenBurn
from xahau.models.transactions.uritoken_buy import URITokenBuy
from xahau.models.transactions.uritoken_cancel_sell_offer import URITokenCancelSellOffer
from xahau.models.transactions.uritoken_create_sell_offer import URITokenCreateSellOffer
from xahau.models.transactions.uritoken_mint import (
    URITokenMint,
    URITokenMintFlag,
    URITokenMintFlagInterface,
)
from xahau.models.transactions.xchain_account_create_commit import (
    XChainAccountCreateCommit,
)
from xahau.models.transactions.xchain_add_account_create_attestation import (
    XChainAddAccountCreateAttestation,
)
from xahau.models.transactions.xchain_add_claim_attestation import (
    XChainAddClaimAttestation,
)
from xahau.models.transactions.xchain_claim import XChainClaim
from xahau.models.transactions.xchain_commit import XChainCommit
from xahau.models.transactions.xchain_create_bridge import XChainCreateBridge
from xahau.models.transactions.xchain_create_claim_id import XChainCreateClaimID
from xahau.models.transactions.xchain_modify_bridge import (
    XChainModifyBridge,
    XChainModifyBridgeFlag,
    XChainModifyBridgeFlagInterface,
)

__all__ = [
    "AccountDelete",
    "AccountSet",
    "AccountSetAsfFlag",
    "AccountSetFlag",
    "AccountSetFlagInterface",
    "AMMBid",
    "AMMCreate",
    "AMMDelete",
    "AMMDeposit",
    "AMMDepositFlag",
    "AMMDepositFlagInterface",
    "AMMVote",
    "AMMWithdraw",
    "AMMWithdrawFlag",
    "AMMWithdrawFlagInterface",
    "AuthAccount",
    "CheckCancel",
    "CheckCash",
    "CheckCreate",
    "ClaimReward",
    "Clawback",
    "DepositPreauth",
    "DIDDelete",
    "DIDSet",
    "EscrowCancel",
    "EscrowCreate",
    "EscrowFinish",
    "Import",
    "Invoke",
    "Memo",
    "NFTokenAcceptOffer",
    "NFTokenBurn",
    "NFTokenCancelOffer",
    "NFTokenCreateOffer",
    "NFTokenCreateOfferFlag",
    "NFTokenCreateOfferFlagInterface",
    "NFTokenMint",
    "NFTokenMintFlag",
    "NFTokenMintFlagInterface",
    "OfferCancel",
    "OfferCreate",
    "OfferCreateFlag",
    "OfferCreateFlagInterface",
    "OracleDelete",
    "OracleSet",
    "Payment",
    "PaymentChannelClaim",
    "PaymentChannelClaimFlag",
    "PaymentChannelClaimFlagInterface",
    "PaymentChannelCreate",
    "PaymentChannelFund",
    "PaymentFlag",
    "PaymentFlagInterface",
    "Remit",
    "MintURIToken",
    "HookGrant",
    "HookParameter",
    "Hook",
    "SetHook",
    "SetHookFlag",
    "SetHookFlagInterface",
    "SetRegularKey",
    "Signer",
    "SignerEntry",
    "SignerListSet",
    "TicketCreate",
    "Transaction",
    "TransactionMetadata",
    "TrustSet",
    "TrustSetFlag",
    "TrustSetFlagInterface",
    "URITokenBurn",
    "URITokenBuy",
    "URITokenCancelSellOffer",
    "URITokenCreateSellOffer",
    "URITokenMint",
    "URITokenMintFlag",
    "URITokenMintFlagInterface",
    "XChainAccountCreateCommit",
    "XChainAddAccountCreateAttestation",
    "XChainAddClaimAttestation",
    "XChainClaim",
    "XChainCommit",
    "XChainCreateBridge",
    "XChainCreateClaimID",
    "XChainModifyBridge",
    "XChainModifyBridgeFlag",
    "XChainModifyBridgeFlagInterface",
]



"""High-level ledger methods with the XAHL ledger."""

from typing import Optional, cast

from xahau.asyncio.clients import Client, XAHLRequestFailureException
from xahau.asyncio.ledger.utils import calculate_fee_dynamically
from xahau.constants import XAHLException
from xahau.models.requests import Fee, GenericRequest, Ledger
from xahau.utils import xah_to_drops


async def get_latest_validated_ledger_sequence(client: Client) -> int:
    """
    Returns the sequence number of the latest validated ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest validated ledger.

    Raises:
        XAHLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Ledger(ledger_index="validated"))
    if response.is_successful():
        return cast(int, response.result["ledger_index"])

    raise XAHLRequestFailureException(response.result)


async def get_latest_open_ledger_sequence(client: Client) -> int:
    """
    Returns the sequence number of the latest open ledger.

    Args:
        client: The network client to use to send the request.

    Returns:
        The sequence number of the latest open ledger.

    Raises:
        XAHLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Ledger(ledger_index="open"))
    if response.is_successful():
        return cast(int, response.result["ledger_index"])

    raise XAHLRequestFailureException(response.result)


async def get_fee(
    client: Client, *, max_fee: Optional[float] = 2, fee_type: str = "open"
) -> str:
    """
    Query the ledger for the current transaction fee.

    Args:
        client: the network client used to make network calls.
        max_fee: The maximum fee in XAH that the user wants to pay. If load gets too
            high, then the fees will not scale past the maximum fee. If None, there is
            no ceiling for the fee. The default is 2 XAH.
        fee_type: The type of fee to return. The options are "open" (the load-scaled
            fee to get into the open ledger), "minimum" (the minimum transaction
            fee) or "dynamic" (dynamic fee-calculation based on the queue size
            of the node). The default is "open". The recommended option is
            "dynamic".

    Returns:
        The transaction fee, in drops.
        `Read more about drops <https://xrpl.org/currency-formats.html#xrp-amounts>`_

    Raises:
        XAHLException: if an incorrect option for `fee_type` is passed in.
        XAHLRequestFailureException: if the rippled API call fails.
    """
    response = await client._request_impl(Fee())
    if not response.is_successful():
        raise XAHLRequestFailureException(response.result)

    result = response.result
    drops = result["drops"]
    if fee_type == "open":
        fee = cast(str, drops["open_ledger_fee"])
    elif fee_type == "minimum":
        fee = cast(str, drops["minimum_fee"])
    elif fee_type == "dynamic":
        fee = calculate_fee_dynamically(fee_data_set=result)
    else:
        raise XAHLException(
            '`fee_type` param must be "open", "minimum" or "dynamic".'
            f" {fee_type} is not a valid option."
        )
    if max_fee is not None:
        max_fee_drops = int(xah_to_drops(max_fee))
        if max_fee_drops < int(fee):
            fee = str(max_fee_drops)
    return fee


async def get_fee_estimate(client: Client, tx_blob: str) -> str:
    """
    Query the ledger for the estimated transaction fee.

    Args:
        client: the network client used to make network calls.
        tx_blob: the encoded transaction that you want the fee estimate for.

    Returns:
        The transaction fee, in drops.
        `Read more about drops <https://xrpl.org/currency-formats.html#xrp-amounts>`_

    Raises:
        XAHLException: if an incorrect option for `fee_type` is passed in.
        XAHLRequestFailureException: if the rippled API call fails.
    """
    fee_request = GenericRequest(command="fee", tx_blob=tx_blob)
    response = await client._request_impl(fee_request)
    if not response.is_successful():
        raise XAHLRequestFailureException(response.result)

    result = response.result
    if "drops" not in result or "base_fee" not in result["drops"]:
        raise XAHLException("could not estimate transaction fee")

    return str(result["drops"]["base_fee"])



"""Transaction and pseudo-transaction type Enums."""

from xahau.models.transactions.types.pseudo_transaction_type import (
    PseudoTransactionType,
)
from xahau.models.transactions.types.transaction_type import TransactionType

__all__ = ["PseudoTransactionType", "TransactionType"]



"""
Specifies an amount in an issued currency.

See https://xrpl.org/currency-formats.html#issued-currency-amounts.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

from xahau.models.amounts.issued_currency_amount import IssuedCurrencyAmount
from xahau.models.base_model import BaseModel
from xahau.models.required import REQUIRED
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class InnerAmount(BaseModel):
    """Represents an amount entry object."""

    amount: Union[IssuedCurrencyAmount, str] = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class AmountEntry(BaseModel):
    """Represents an amount entry object."""

    amount_entry: InnerAmount = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """


Amounts = List[AmountEntry]



"""Model for ClaimReward transaction type."""

from dataclasses import dataclass, field
from typing import Optional

from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class ClaimReward(Transaction):
    """
    Represents a `ClaimReward <https://xrpl.org/claimreward.html>`_ transaction,
    which triggers the reward hook. You can also use this trasaction to opt out
    of rewards
    """

    issuer: Optional[str] = None
    """
    The address of the `issuer
    <https://xrpl.org/accounts.html>`_ where the reward.c hook is installed.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.CLAIM_REWARD,
        init=False,
    )



"""
The XAH Ledger has two kinds of money: XAH, and issued
currencies. Both types have high precision, although their
formats are different.
"""
from typing import Union

from xahau.models.currencies.issued_currency import IssuedCurrency
from xahau.models.currencies.xah import XAH

Currency = Union[IssuedCurrency, XAH]



"""Model for URITokenMint transaction type."""

from dataclasses import dataclass, field

from xahau.models.amounts import Amount
from xahau.models.required import REQUIRED
from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class URITokenBuy(Transaction):
    """
    The URITokenBuy transaction is used to purchase a token for the sell offer
    Amount.
    """

    uritoken_id: str = REQUIRED  # type: ignore
    """
    Identifies the TokenID of the URIToken object that the
    offer references. This field is required.

    :meta hide-value:
    """

    amount: Amount = REQUIRED  # type: ignore
    """
    Indicates the amount expected or offered for the Token.

    The amount must be non-zero, except when this is a sell
    offer and the asset is XAH. This would indicate that the current
    owner of the token is giving it away free, either to anyone at all,
    or to the account identified by the Destination field. This field
    is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.URITOKEN_BUY,
        init=False,
    )



"""Model for SetHook transaction type."""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Pattern, Union

from typing_extensions import Final

from xahau.models.flags import FlagInterface
from xahau.models.nested_model import NestedModel
from xahau.models.required import REQUIRED
from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init

MAX_HOOKS: Final[int] = 10
"""
Maximum number of hooks allowed.

:meta private:
"""
HEX_REGEX: Final[Pattern[str]] = re.compile("[A-Fa-f0-9]{64}")
"""
Matches hex-encoded WalletLocator in the format allowed by XAHL.

:meta private:
"""


class SetHookFlag(int, Enum):
    """SetHookFlag"""

    HSF_OVERRIDE = 0x00000001
    """"""

    HSF_NS_DELETE = 0x00000002
    """"""

    HSF_COLLECT = 0x00000004
    """"""


class SetHookFlagInterface(FlagInterface):
    """SetHookFlagInterface"""

    HSF_OVERRIDE: bool
    HSF_NS_DELETE: bool
    HSF_COLLECT: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class HookGrant(NestedModel):
    """Represents one grant in a list of grants on the transaction."""

    hook_hash: str = REQUIRED  # type: ignore
    """
    The hook hash of the grant.
    This field is required.

    :meta hide-value:
    """

    authorize: Optional[str] = None
    """The account authorized on the grant."""


@require_kwargs_on_init
@dataclass(frozen=True)
class HookParameter(NestedModel):
    """Represents one parameter in a list of parameters on the transaction."""

    hook_parameter_name: str = REQUIRED  # type: ignore
    """
    The name of the parameter.
    This field is required.

    :meta hide-value:
    """

    hook_parameter_value: str = REQUIRED  # type: ignore
    """
    The value of the parameter.
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class Hook(NestedModel):
    """Represents one hook in a list of hooks deployed to an account."""

    create_code: Optional[str] = None
    """
    The code that is executed when the hook is triggered.
    This field is required.

    :meta hide-value:
    """

    flags: Union[Dict[str, bool], int, List[int]] = 0
    """
    The flags that are set on the hook.
    This field is required.

    :meta hide-value:
    """

    hook_hash: Optional[str] = None
    """"""

    hook_on: Optional[str] = None
    """The transactions that triggers the hook. Represented as a 256Hash"""

    hook_namespace: Optional[str] = None
    """The namespace of the hook."""

    hook_api_version: Optional[int] = None
    """The API version of the hook."""

    hook_parameters: Optional[List[HookParameter]] = None
    """The parameters of the hook."""

    hook_grants: Optional[List[HookGrant]] = None
    """The grants of the hook."""


@require_kwargs_on_init
@dataclass(frozen=True)
class SetHook(Transaction):
    """Sets the an array of hooks on an account."""

    hooks: Optional[List[Hook]] = None
    transaction_type: TransactionType = field(
        default=TransactionType.SET_HOOK,
        init=False,
    )

    def _get_errors(self: SetHook) -> Dict[str, str]:
        errors = super()._get_errors()

        if self.hooks is None:  # deletion of the Hooks object
            return errors

        if len(self.hooks) > MAX_HOOKS:
            errors["hooks"] = "`hooks` must have no more than {} members.".format(
                MAX_HOOKS
            )
            return errors

        for hook in self.hooks:
            if hook.hook_on is not None and not bool(HEX_REGEX.fullmatch(hook.hook_on)):
                errors["hooks"] = "A Hook's hook_on must be a 256-bit (32-byte) "
                "hexadecimal value."

            if hook.hook_namespace is not None and not bool(
                HEX_REGEX.fullmatch(hook.hook_namespace)
            ):
                errors["hooks"] = "A Hook's hook_namespace must be a 256-bit (32-byte) "
                "hexadecimal value."

        return errors



"""Model for SetHook transaction type."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, Optional

from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class Import(Transaction):
    """Imorts an xpop into the account."""

    issuer: Optional[str] = None
    """
    If present, invokes the Hook on the Issuer account.
    """

    blob: Optional[str] = None
    """
    Hex value representing a VL Blob.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.IMPORT,
        init=False,
    )

    def _get_errors(self: Import) -> Dict[str, str]:
        errors = super()._get_errors()
        return errors



"""
The ledger_closed method returns the unique
identifiers of the most recently closed ledger.
(This ledger is not necessarily validated and
immutable yet.)
"""
from dataclasses import dataclass, field

from xahau.models.requests.request import Request, RequestMethod
from xahau.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class LedgerClosed(Request):
    """
    The ledger_closed method returns the unique
    identifiers of the most recently closed ledger.
    (This ledger is not necessarily validated and
    immutable yet.)
    """

    method: RequestMethod = field(default=RequestMethod.LEDGER_CLOSED, init=False)



"""
This request retrieves the namespace for an account.

All information retrieved is relative to a particular version of the ledger.

`See account_info <https://xrpl.org/account_namespace.html>`_
"""
from dataclasses import dataclass, field

from xahau.models.requests.request import Request, RequestMethod
from xahau.models.required import REQUIRED
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class AccountNamespace(Request):
    """
    This request retrieves the namespace for an account.

    All information retrieved is relative to a particular version of the ledger.

    `See account_info <https://xrpl.org/account_namespace.html>`_
    """

    account: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    namespace_id: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.ACCOUNT_NAMESPACE, init=False)



"""High-level XAHL exports."""
from xahau import account, clients, core, ledger, models, transaction, utils, wallet
from xahau.constants import CryptoAlgorithm, XAHLException

__all__ = [
    "CryptoAlgorithm",
    "XAHLException",
    "account",
    "clients",
    "core",
    "ledger",
    "models",
    "transaction",
    "utils",
    "wallet",
]



"""Hash Utilities"""

import hashlib

from xahau.core import addresscodec
from xahau.utils import str_to_hex

# Constants
HEX = 16
BYTE_LENGTH = 32

# Ledger space dictionary
ledger_spaces = {
    "account": "a",
    "dirNode": "d",
    "generatorMap": "g",
    "rippleState": "r",
    "offer": "o",
    "ownerDir": "O",
    "bookDir": "B",
    "contract": "c",
    "skipList": "s",
    "escrow": "u",
    "amendment": "f",
    "feeSettings": "e",
    "ticket": "T",
    "signerList": "S",
    "paychan": "x",
    "check": "C",
    "uriToken": "U",
    "depositPreauth": "p",
}


def sha512_half(data: str) -> str:
    """Compute the SHA-512 hash and then take the first half of the result.

    Args:
        data (str): The input data in hexadecimal format.

    Returns:
        str: The first half of the SHA-512 hash in uppercase hexadecimal.
    """
    hash_obj = hashlib.sha512(bytes.fromhex(data))
    return hash_obj.hexdigest()[:64].upper()


def address_to_hex(address: str) -> str:
    """Convert an address to its hexadecimal representation.

    Args:
        address (str): The address to convert.

    Returns:
        str: The hexadecimal representation of the address.
    """
    return addresscodec.decode_classic_address(address).hex()


def ledger_space_hex(name: str) -> str:
    """Get the hexadecimal representation of a ledger space.

    Args:
        name (str): The name of the ledger space.

    Returns:
        str: The hexadecimal representation of the ledger space.
    """
    return format(ord(ledger_spaces[name]), "x").zfill(4)


def hash_offer(address: str, sequence: int) -> str:
    """Compute the hash of an Offer.

    Args:
        address (str): The address associated with the offer.
        sequence (int): The sequence number of the offer.

    Returns:
        str: The computed hash of the offer in uppercase hexadecimal.
    """
    return sha512_half(
        ledger_space_hex("offer")
        + address_to_hex(address)
        + format(sequence, "x").zfill(BYTE_LENGTH * 2)
    ).upper()


def hash_check(address: str, sequence: int) -> str:
    """Compute the hash of a Check.

    Args:
        address (str): The address associated with the check.
        sequence (int): The sequence number of the check.

    Returns:
        str: The computed hash of the check in uppercase hexadecimal.
    """
    return sha512_half(
        ledger_space_hex("check")
        + address_to_hex(address)
        + format(sequence, "x").zfill(BYTE_LENGTH * 2)
    ).upper()


def hash_escrow(address: str, sequence: int) -> str:
    """Compute the hash of an Escrow.

    Args:
        address (str): The address associated with the escrow.
        sequence (int): The sequence number of the escrow.

    Returns:
        str: The computed hash of the escrow in uppercase hexadecimal.
    """
    return sha512_half(
        ledger_space_hex("escrow")
        + address_to_hex(address)
        + format(sequence, "x").zfill(BYTE_LENGTH * 2)
    ).upper()


def hash_payment_channel(address: str, dst_address: str, sequence: int) -> str:
    """Compute the hash of a Payment Channel.

    Args:
        address (str): The address of the payment channel.
        dst_address (str): The destination address for the payment channel.
        sequence (int): The sequence number of the payment channel.

    Returns:
        str: The computed hash of the payment channel in uppercase hexadecimal.
    """
    return sha512_half(
        ledger_space_hex("paychan")
        + address_to_hex(address)
        + address_to_hex(dst_address)
        + format(sequence, "x").zfill(BYTE_LENGTH * 2)
    ).upper()


def hash_uri_token(issuer: str, uri: str) -> str:
    """Compute the hash of a URIToken.

    Args:
        issuer (str): The address of the issuer.
        uri (str): The URI associated with the token.

    Returns:
        str: The computed hash of the URI token in uppercase hexadecimal.
    """
    return sha512_half(
        ledger_space_hex("uriToken") + address_to_hex(issuer) + str_to_hex(uri)
    ).upper()


def hash_hook(hex: str) -> str:
    """Compute the hash of a Hook.

    Args:
        hex (str): The hexadecimal representation of the hook.

    Returns:
        str: The computed hash of the hook in uppercase hexadecimal.
    """
    return sha512_half(hex).upper()



"""Model for URITokenMint transaction type."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Optional

from typing_extensions import Final

from xahau.models.amounts import Amount
from xahau.models.flags import FlagInterface
from xahau.models.required import REQUIRED
from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init

_MAX_URI_LENGTH: Final[int] = 512
_MAX_DIGEST_LENGTH: Final[int] = 64


class URITokenMintFlag(int, Enum):
    """
    Transactions of the URITokenMint type support additional values in the Flags
    field. This enum represents those options.

    `See URITokenMint Flags
    <https://xrpl.org/uritokenmint.html#uritokenmint-flags>`_
    """

    TF_BURNABLE = 0x00000001
    """
    If set, indicates that the minted token may be burned by the issuer even
    if the issuer does not currently hold the token. The current holder of
    the token may always burn it.
    """


class URITokenMintFlagInterface(FlagInterface):
    """
    Transactions of the URITokenMint type support additional values in the Flags
    field. This TypedDict represents those options.

    `See URITokenMint Flags
    <https://xrpl.org/uritokenmint.html#uritokenmint-flags>`_
    """

    TF_BURNABLE: bool


@require_kwargs_on_init
@dataclass(frozen=True)
class URITokenMint(Transaction):
    """The URITokenMint transaction creates an URIToken object."""

    uri: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    digest: Optional[str] = None
    """
    """

    destination: Optional[str] = None
    """
    If present, indicates that this offer may only be
    accepted by the specified account. Attempts by other
    accounts to accept this offer MUST fail.
    """

    amount: Optional[Amount] = None
    """
    Indicates the amount expected or offered for the Token.

    The amount must be non-zero, except when this is a sell
    offer and the asset is XAH. This would indicate that the current
    owner of the token is giving it away free, either to anyone at all,
    or to the account identified by the Destination field.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.URITOKEN_MINT,
        init=False,
    )

    def _get_errors(self: "URITokenMint") -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "uri": self._get_uri_error(),
                "amount": self._get_amount_error(),
                "destination": self._get_destination_error(),
                "digest": self._get_digest_error(),
            }.items()
            if value is not None
        }

    def _get_uri_error(self: "URITokenMint") -> Optional[str]:
        if self.uri is not None and len(self.uri) > _MAX_URI_LENGTH:
            return f"Must not be longer than {_MAX_URI_LENGTH} characters"
        return None

    def _get_amount_error(self: "URITokenMint") -> Optional[str]:
        if self.amount == 0 and not self.destination:
            return "Cannot be 0 without destination"
        return None

    def _get_destination_error(self: "URITokenMint") -> Optional[str]:
        if self.destination == self.account:
            return "Must not be equal to the account"
        return None

    def _get_digest_error(self: "URITokenMint") -> Optional[str]:
        if self.digest is not None and len(self.digest) > _MAX_DIGEST_LENGTH:
            return f"Must not be longer than {_MAX_DIGEST_LENGTH} characters"
        return None



"""
The XAH Ledger has two kinds of money: XAH, and issued
currencies. Both types have high precision, although their
formats are different.
"""
from xahau.models.currencies.currency import Currency
from xahau.models.currencies.issued_currency import IssuedCurrency
from xahau.models.currencies.xah import XAH

__all__ = [
    "Currency",
    "IssuedCurrency",
    "XAH",
]



"""Model for URITokenBurn transaction type."""
from dataclasses import dataclass, field

from xahau.models.required import REQUIRED
from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True)
class URITokenBurn(Transaction):
    """
    The URITokenBurn transaction is used to remove an URIToken object from the
    Account Objects in which it is being held, effectively removing the token from
    the ledger ("burning" it).
    """

    account: str = REQUIRED  # type: ignore
    """
    Identifies the AccountID that submitted this transaction. The account must
    be the present owner of the token or, if the lsfBurnable flag is set
    on the URIToken, either the issuer account or an account authorized by the
    issuer (i.e. MintAccount). This field is required.

    :meta hide-value:
    """

    uritoken_id: str = REQUIRED  # type: ignore
    """
    Identifies the URIToken to be burned. This field is required.

    :meta hide-value:
    """

    transaction_type: TransactionType = field(
        default=TransactionType.URITOKEN_BURN,
        init=False,
    )



"""Top-level exports for types used in binary_codec."""
from xahau.core.binarycodec.types.account_id import AccountID
from xahau.core.binarycodec.types.amount import Amount
from xahau.core.binarycodec.types.blob import Blob
from xahau.core.binarycodec.types.currency import Currency
from xahau.core.binarycodec.types.hash import Hash
from xahau.core.binarycodec.types.hash128 import Hash128
from xahau.core.binarycodec.types.hash160 import Hash160
from xahau.core.binarycodec.types.hash256 import Hash256
from xahau.core.binarycodec.types.issue import Issue
from xahau.core.binarycodec.types.path_set import PathSet
from xahau.core.binarycodec.types.st_array import STArray
from xahau.core.binarycodec.types.st_object import STObject
from xahau.core.binarycodec.types.uint import UInt
from xahau.core.binarycodec.types.uint8 import UInt8
from xahau.core.binarycodec.types.uint16 import UInt16
from xahau.core.binarycodec.types.uint32 import UInt32
from xahau.core.binarycodec.types.uint64 import UInt64
from xahau.core.binarycodec.types.vector256 import Vector256
from xahau.core.binarycodec.types.xchain_bridge import XChainBridge

__all__ = [
    "AccountID",
    "Amount",
    "Blob",
    "Currency",
    "Hash",
    "Hash128",
    "Hash160",
    "Hash256",
    "Issue",
    "PathSet",
    "STObject",
    "STArray",
    "UInt",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "Vector256",
    "XChainBridge",
]



"""
The book_offers method retrieves a list of offers, also known
as the order book, between two currencies.
"""
from dataclasses import dataclass, field
from typing import Optional

from xahau.models.currencies import Currency
from xahau.models.requests.request import LookupByLedgerRequest, Request, RequestMethod
from xahau.models.required import REQUIRED
from xahau.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class BookOffers(Request, LookupByLedgerRequest):
    """
    The book_offers method retrieves a list of offers, also known
    as the order book, between two currencies.
    """

    taker_gets: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    taker_pays: Currency = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    method: RequestMethod = field(default=RequestMethod.BOOK_OFFERS, init=False)
    limit: Optional[int] = None
    taker: Optional[str] = None



"""Async methods for working with XAHL wallets."""
from xahau.asyncio.wallet.wallet_generation import (
    XAHLFaucetException,
    generate_faucet_wallet,
)

__all__ = ["XAHLFaucetException", "generate_faucet_wallet"]



"""Model for URITokenMint transaction type."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union

from typing_extensions import Final

from xahau.models.amounts import AmountEntry
from xahau.models.base_model import BaseModel
from xahau.models.required import REQUIRED
from xahau.models.transactions.transaction import Transaction
from xahau.models.transactions.types import TransactionType
from xahau.models.utils import require_kwargs_on_init

_MAX_BLOB_LENGTH: Final[int] = 256
_MAX_URI_LENGTH: Final[int] = 512
_DIGEST_LENGTH: Final[int] = 64


@require_kwargs_on_init
@dataclass(frozen=True)
class MintURIToken(BaseModel):
    """Represents a uritoken object."""

    uri: str = REQUIRED  # type: ignore
    """
    This field is required.

    :meta hide-value:
    """

    digest: Optional[str] = None
    """
    """

    flags: Union[Dict[str, bool], int, List[int]] = 0
    """
    The flags that are set on the uri token.
    This field is required.

    :meta hide-value:
    """


@require_kwargs_on_init
@dataclass(frozen=True)
class Remit(Transaction):
    """The Remit transaction"""

    destination: str = REQUIRED  # type: ignore
    """
    The address of the account receiving the payment. This field is required.

    :meta hide-value:
    """

    destination_tag: Optional[int] = None
    """
    An arbitrary `destination tag
    <https://xrpl.org/source-and-destination-tags.html>`_ that
    identifies the reason for the Payment, or a hosted recipient to pay.
    """

    amounts: Optional[List[AmountEntry]] = None

    mint_uri_token: Optional[MintURIToken] = None

    uri_token_ids: Optional[List[str]] = None

    invoice_id: Optional[str] = None  # TODO: should be a 256 bit hash
    """
    Arbitrary 256-bit hash representing a specific reason or identifier for
    this Check.
    """

    blob: Optional[str] = None
    """
    Hex value representing a VL Blob.
    """

    inform: Optional[str] = None
    """
    If present, can trigger a hook on a third party account.
    """

    transaction_type: TransactionType = field(
        default=TransactionType.REMIT,
        init=False,
    )

    def _get_errors(self: "Remit") -> Dict[str, str]:
        return {
            key: value
            for key, value in {
                **super()._get_errors(),
                "destination": self._get_destination_error(),
                "inform": self._get_inform_error(),
                "amounts": self._get_amounts_error(),
                "uri_token_ids": self._get_uri_token_ids_error(),
                "uri": self._get_uri_error(),
                "digest": self._get_digest_error(),
                "blob": self._get_blob_error(),
            }.items()
            if value is not None
        }

    def _get_amounts_error(self: "Remit") -> Optional[str]:
        if self.amounts is not None and len(self.amounts) < 1:
            return "Empty Amounts"
        if self.amounts is not None and len(self.amounts) > 32:
            return "Max Amounts"
        if self.amounts is not None:
            seen = set()
            seen_xrp = False
            for amount in self.amounts:
                if isinstance(amount.amount_entry.amount, str):
                    if seen_xrp:
                        return "Duplicate Native amounts are not allowed"
                    seen_xrp = True
                else:
                    amount_key = (
                        amount.amount_entry.amount.currency,
                        amount.amount_entry.amount.issuer,
                    )
                    if amount_key in seen:
                        return "Duplicate amounts are not allowed"
                    seen.add(amount_key)
        return None

    def _get_uri_token_ids_error(self: "Remit") -> Optional[str]:
        if self.uri_token_ids is not None and len(self.uri_token_ids) < 1:
            return "Empty URI token IDs"
        if self.uri_token_ids is not None and len(self.uri_token_ids) > 32:
            return "Max URI token IDs"
        if self.uri_token_ids is not None:
            seen = set()
            for token_id in self.uri_token_ids:
                if not isinstance(token_id, str) or len(token_id) != 64:
                    return "Each URI token ID must be a 64-character hash256 string"
                if token_id in seen:
                    return "Duplicate URI token IDs are not allowed"
                seen.add(token_id)
        return None

    def _get_destination_error(self: "Remit") -> Optional[str]:
        if self.destination == self.account:
            return "Must not be equal to the account"
        return None

    def _get_inform_error(self: "Remit") -> Optional[str]:
        if self.inform == self.account or self.inform == self.destination:
            return "Must not be equal to the account or destination"
        return None

    def _get_uri_error(self: "Remit") -> Optional[str]:
        if (
            self.mint_uri_token is not None
            and len(self.mint_uri_token.uri) > _MAX_URI_LENGTH
        ):
            return f"Must not be longer than {_MAX_URI_LENGTH} characters"
        return None

    def _get_digest_error(self: "Remit") -> Optional[str]:
        if (
            self.mint_uri_token is not None
            and self.mint_uri_token.digest is not None
            and len(self.mint_uri_token.digest) != _DIGEST_LENGTH
        ):
            return f"Must be exactly {_DIGEST_LENGTH} characters"
        return None

    def _get_blob_error(self: "Remit") -> Optional[str]:
        if self.blob is not None and len(self.blob) > _MAX_BLOB_LENGTH:
            return f"Must not be longer than {_MAX_BLOB_LENGTH} characters"
        return None



"""
Specifies XAH as a currency, without a value. Normally, you will not use this
model as it does not specify an amount of XAH. In cases where you need to
specify an amount of XAH, you will use a string. However, for some book order
requests where currencies are specified without amounts, you may need to
specify the use of XAH, without a value. In these cases, you will use this
object.

See https://xrpl.org/currency-formats.html#specifying-currency-amounts
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Type, Union

from typing_extensions import Self

from xahau.models.base_model import BaseModel
from xahau.models.exceptions import XAHLModelException
from xahau.models.utils import KW_ONLY_DATACLASS, require_kwargs_on_init


@require_kwargs_on_init
@dataclass(frozen=True, **KW_ONLY_DATACLASS)
class XAH(BaseModel):
    """
    Specifies XAH as a currency, without a value. Normally, you will not use this
    model as it does not specify an amount of XAH. In cases where you need to
    specify an amount of XAH, you will use a string. However, for some book order
    requests where currencies are specified without amounts, you may need to
    specify the use of XAH, without a value. In these cases, you will use this
    object.

    See https://xrpl.org/currency-formats.html#specifying-currency-amounts
    """

    currency: str = field(default="XAH", init=False)

    @classmethod
    def from_dict(cls: Type[Self], value: Dict[str, Any]) -> Self:
        """
        Construct a new XAH from a dictionary of parameters.

        Args:
            value: The value to construct the XAH from.

        Returns:
            A new XAH object, constructed using the given parameters.

        Raises:
            XAHLModelException: If the dictionary provided is invalid.
        """
        if len(value) != 1 or "currency" not in value or value["currency"] != "XAH":
            raise XAHLModelException("Not a valid XAH type")
        return cls()

    def to_dict(self: Self) -> Dict[str, Any]:
        """
        Returns the dictionary representation of an XAH currency object.

        Returns:
            The dictionary representation of an XAH currency object.
        """
        return {**super().to_dict(), "currency": "XAH"}

    def to_amount(self: Self, value: Union[str, int, float]) -> str:
        """
        Converts value to XAH.

        Args:
            value: The amount of XAH.

        Returns:
            A string representation of XAH amount.
        """
        # import needed here to avoid circular dependency
        from xahau.utils.xah_conversions import xah_to_drops

        if isinstance(value, str):
            return xah_to_drops(float(value))
        return xah_to_drops(value)

    def __repr__(self: Self) -> str:
        """
        Generate string representation of XAH.

        Returns:
            A string representation of XAH currency.
        """
        return "XAH()"



#!/usr/bin/env python
# coding: utf-8

"""Hooks-related helper util functions."""

import binascii
from typing import Any, Dict, List  # noqa: F401

from xahau.constants import XAHLException
from xahau.core.binarycodec.definitions import _TRANSACTION_TYPE_MAP, _TRANSACTION_TYPES


def calculate_hook_on(arr: List[str]) -> str:
    """
    Calculate the hook on value for a given list of transaction types.

    Args:
        arr: List of transaction types.

    Returns:
        A 256 hash of the transactions the hook will invoke on

    Raises:
        XAHLException: if the HookOn transaction type is not in the transaction types
    """
    tts = _TRANSACTION_TYPE_MAP
    s = "0xffffffffffffffffffffffffffffffffffffffffffffffffffffffffffbfffff"
    for n in arr:
        if n not in _TRANSACTION_TYPES:
            raise XAHLException(f"invalid transaction type '{n}' in HookOn array")
        v = int(s, 16)
        v ^= 1 << tts[n]
        s = "0x" + hex(v)[2:]

    s = s.replace("0x", "")
    s = s.zfill(64)
    return s.upper()


def hex_hook_parameters(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Hexlify the hook parameters

    Args:
        data: A list of Hook Parameters

    Returns:
        A list of Hook Parameters with the values in hex format
    """
    hook_parameters: List[Dict[str, Any]] = []
    for parameter in data:
        hook_parameters.append(
            {
                "HookParameter": {
                    "HookParameterName": binascii.hexlify(
                        parameter["HookParameter"]["HookParameterName"].encode("utf8")
                    )
                    .decode("utf-8")
                    .upper(),
                    "HookParameterValue": binascii.hexlify(
                        parameter["HookParameter"]["HookParameterValue"].encode("utf8")
                    )
                    .decode("utf-8")
                    .upper(),
                }
            }
        )
    return hook_parameters
